# Copyright (C) 2024 Mitsubishi Electric Research Laboratories (MERL)
#
# SPDX-License-Identifier: AGPL-3.0-or-later

# distutils: language = c++
#cython: language_level=3

import numpy as np

cimport cython
cimport numpy as np
from libcpp.vector cimport vector
from numpy.math cimport INFINITY


def change_detection(sed_scores, timestamps_in_sec, filter_length_in_sec, time_decimals=6):
    """change point detection in the piecewise constant SED posterior scores by
    employing a time continuous step filter.

    Args:
        sed_scores (2d np.ndarray): Predicted SED posterior scores of shape NxC,
            with N being the number of frames and C the number of sound classes.
        timestamps_in_sec (1d np.ndarray): The N+1 frame boundary times in sec
        filter_length_in_sec: filter length in sec T_fil. The step filter equals
            -2/T_fil \in [0,T_fil/2] (first half of the filter) and
            +2/T_fil \in [T_fil/2,T_fil] (second half of the filter). Hence, at
            each time point the filter predicts change/delta values as the
            difference between the average score in the next T_fil/2 seconds and
            the previous T_fil/2 seconds.
        time_decimals: the decimal precision to be used for timestamps.

    Returns (list of tuple): for each sound class:
        seg_bounds (1d np.ndarray): segment boundaries/change points.
            seg_bounds[0] equals audio onset 0., seg_bounds[1:-2:2] are event
            onset candidates, seg_bounds[2:-1:2] are event offset candidates,
            and seg_bounds[-1] equals audio length.
        mean_scores (1d np.ndarray): the average score within each segment.
        min_scores (1d np.ndarray): the minimum score within each segment.
        max_scores (1d np.ndarray): the maximum score within each segment.

    """
    assert filter_length_in_sec is not None, "You must provide filter_length_in_sec"

    if np.isscalar(filter_length_in_sec):
        filter_length_in_sec = np.array(sed_scores.shape[1] * [filter_length_in_sec], dtype=np.float64)

    sed_scores = np.asanyarray(sed_scores, dtype=np.float64)
    assert sed_scores.ndim == 2, sed_scores.shape
    timestamps_in_sec = np.asanyarray(timestamps_in_sec, dtype=np.float64)
    assert timestamps_in_sec.ndim == 1, timestamps_in_sec.shape
    assert (timestamps_in_sec[1:] > timestamps_in_sec[:-1]).all(), np.min(timestamps_in_sec[1:] - timestamps_in_sec[:-1])
    filter_length_in_sec = np.asanyarray(filter_length_in_sec, dtype=np.float64)
    cdef int num_timesteps = timestamps_in_sec.shape[0] + 2
    # breakpoint times, where the filter center, onset or offset moves into a new frame:
    cdef double [:,:] overlap_breakpoint_timestamps = np.array([
        np.sort(
            np.round(np.concatenate((
                timestamps_in_sec,
                timestamps_in_sec + fil_len / 2,
                timestamps_in_sec - fil_len / 2,
            )), decimals=time_decimals+2)
        )
        for fil_len in filter_length_in_sec
    ])
    # filter output is piecewise linear between these breakpoints and local min/max do lie on breakpoints.

    cdef double [:] filter_length = filter_length_in_sec

    cdef double [:,:] scores = np.concatenate((
        np.full_like(sed_scores[:1], 0),
        sed_scores,
        np.full_like(sed_scores[:1], 0),
    ))
    cdef double [:] timestamps = np.concatenate(([-np.inf], timestamps_in_sec, [np.inf]))
    cdef int num_classes = sed_scores.shape[1]
    cdef double eps = 10**(-time_decimals-1)

    cdef:
        int i, k, first_time_seg_idx, center_time_seg_idx, last_time_seg_idx, num_segments
        double t, t_diff, fil_len_k, center_time, left_part_sum, right_part_sum

    cdef vector[double] onsets = []
    cdef vector[double] offsets = []
    cdef vector[double] cum_scores = []
    cdef vector[double] min_scores = []
    cdef vector[double] max_scores = []
    cdef vector[double] deltas = []
    cdef vector[double] deltas_rel = []
    cdef vector[int] class_pos = [0]
    cdef double [:] scores_k
    cdef double [:] overlap_breakpoint_timestamps_k
    cdef int num_onsets = 0
    cdef int num_offsets = 0
    cdef int num_deltas = 0

    for k in range(num_classes):
        scores_k = scores[:, k]
        fil_len_k = filter_length[k]
        overlap_breakpoint_timestamps_k = overlap_breakpoint_timestamps[k]
        first_time_seg_idx = 0
        center_time_seg_idx = 1
        center_time = 0.
        left_part_sum = 0.
        last_time_seg_idx = 1
        right_part_sum = 0.
        while timestamps[last_time_seg_idx+1] <= fil_len_k / 2:
            right_part_sum += scores_k[last_time_seg_idx] * (
                timestamps[last_time_seg_idx+1] - timestamps[last_time_seg_idx]
            )
            last_time_seg_idx += 1
        right_part_sum += scores_k[last_time_seg_idx] * (
            fil_len_k / 2 - timestamps[last_time_seg_idx]
        )
        deltas.clear()
        deltas_rel.clear()
        deltas.push_back(2*(right_part_sum - left_part_sum)/fil_len_k)
        deltas_rel.push_back((right_part_sum+1e-6)/(left_part_sum+1e-6))
        num_deltas = 1
        cum_scores.push_back(0.)
        min_scores.push_back(INFINITY)
        max_scores.push_back(0.)
        for i in range(len(overlap_breakpoint_timestamps_k)):
            t = overlap_breakpoint_timestamps_k[i]  # current breakpoint
            # get time difference to previous breakpoint, i.e., the time the filter
            # moved since last breakpoint
            t_diff = t - overlap_breakpoint_timestamps_k[i-1]
            if (t < eps) or (t_diff < eps) or (t > timestamps[num_timesteps-2]):
                continue
            left_part_sum -= t_diff * scores_k[first_time_seg_idx]
            left_part_sum += t_diff * scores_k[center_time_seg_idx]
            right_part_sum -= t_diff * scores_k[center_time_seg_idx]
            right_part_sum += t_diff * scores_k[last_time_seg_idx]
            center_time += t_diff
            deltas.push_back(2*(right_part_sum - left_part_sum)/fil_len_k)
            deltas_rel.push_back((right_part_sum+1e-3)/(left_part_sum+1e-3))
            num_deltas += 1
            if (num_offsets == num_onsets) and (deltas[num_deltas-1] <= deltas[num_deltas-2]) and ((num_deltas == 2) or (deltas[num_deltas-2] > deltas[num_deltas-3])):
                # previous delta was a local maximum and therefore a new onset candidate
                onsets.push_back(overlap_breakpoint_timestamps_k[i-1])
                num_onsets += 1
                cum_scores.push_back(0.)
                min_scores.push_back(INFINITY)
                max_scores.push_back(0.)
            elif (num_onsets > num_offsets) and (deltas[num_deltas-1] > deltas[num_deltas-2]) and (deltas[num_deltas-2] <= deltas[num_deltas-3]):
                # previous delta was a local minimum and therefore a new offset candidate
                offsets.push_back(overlap_breakpoint_timestamps_k[i-1])
                num_offsets += 1
                cum_scores.push_back(0.)
                min_scores.push_back(INFINITY)
                max_scores.push_back(0.)
            num_segments = 1 + num_onsets + num_offsets + k
            cum_scores[num_segments-1] += t_diff * scores_k[center_time_seg_idx]
            min_scores[num_segments-1] = min(min_scores[num_segments-1], scores_k[center_time_seg_idx])
            max_scores[num_segments-1] = max(max_scores[num_segments-1], scores_k[center_time_seg_idx])
            if (timestamps[center_time_seg_idx+1] - center_time) < eps:
                center_time_seg_idx += 1
            if (timestamps[first_time_seg_idx+1] - (center_time - fil_len_k / 2)) < eps:
                first_time_seg_idx += 1
            if (timestamps[last_time_seg_idx+1] - (center_time + fil_len_k / 2)) < eps:
                last_time_seg_idx += 1
        if num_onsets > num_offsets:
            offsets.push_back(timestamps[num_timesteps-2])
            num_offsets += 1
            cum_scores.push_back(0.)
            min_scores.push_back(0.)
            max_scores.push_back(0.)
        class_pos.push_back(num_onsets)

    assert num_onsets == num_offsets, (num_onsets, num_offsets, onsets, offsets)
    onsets_arr = np.array(onsets)
    offsets_arr = np.array(offsets)
    class_pos_arr = np.array(class_pos)
    cum_scores_arr = np.array(cum_scores)
    min_scores_arr = np.array(min_scores)
    max_scores_arr = np.array(max_scores)
    out = []
    for c in range(int(num_classes)):
        onsets_c = onsets_arr[class_pos_arr[c]:class_pos_arr[c+1]]
        offsets_c = offsets_arr[class_pos_arr[c]:class_pos_arr[c+1]]
        seg_bounds_c = np.concatenate(([0], np.stack((onsets_c, offsets_c), axis=1).flatten(), [timestamps_in_sec[-1]]))
        mean_scores_c = cum_scores_arr[2*class_pos_arr[c]+c:2*class_pos_arr[c+1]+c+1]
        seg_lens_c = np.maximum(seg_bounds_c[1:] - seg_bounds_c[:-1], eps)
        mean_scores_c /= seg_lens_c
        min_scores_c = min_scores_arr[2*class_pos_arr[c]+c:2*class_pos_arr[c+1]+c+1]
        min_scores_c[min_scores_c == np.inf] = 0.
        min_scores_c[0] = min_scores_c[-1] = 0.
        max_scores_c = max_scores_arr[2*class_pos_arr[c]+c:2*class_pos_arr[c+1]+c+1]
        out.append((seg_bounds_c, mean_scores_c, min_scores_c, max_scores_c))
    return out



# Copyright (C) 2024 Mitsubishi Electric Research Laboratories (MERL)
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import numpy as np
from sed_scores_eval import collar_based, intersection_based, io

from sebbs import median_filter, package_dir


def test_median_filter_oracle_tuning_psds():
    # load predictions, ground_truth, and audio_durations
    scores = io.read_sed_scores(package_dir / "tests" / "dcase2023task4a" / "baseline2_run1_scores")
    ground_truth = io.read_ground_truth_events(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "ground_truth.tsv"
    )
    ground_truth = {audio_id: ground_truth[audio_id] for audio_id in ground_truth if audio_id in scores}
    scores = {audio_id: scores[audio_id] for audio_id in ground_truth}
    audio_durations = io.read_audio_durations(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "audio_durations.tsv"
    )
    audio_durations = {audio_id: audio_durations[audio_id] for audio_id in scores}

    # tune hyper-parameters on subset of DESED's public eval
    predictor, best_psds_values = median_filter.tune(
        scores=scores,
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        filter_lengths=np.linspace(0.0, 2.0, 11),
        selection_fn=median_filter.select_best_psds,
    )
    # run inference on subset of DESED's public eval
    median_filtered_scores = predictor.predict(scores)
    # evaluate PSDS performance
    psds, single_class_psds, *_ = intersection_based.psds(
        scores=median_filtered_scores,
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        dtc_threshold=0.7,
        gtc_threshold=0.7,
        alpha_ct=0.0,
        alpha_st=1.0,
        unit_of_time="hour",
        max_efpr=100.0,
    )
    for sound_class in single_class_psds:
        assert abs(single_class_psds[sound_class] - best_psds_values[sound_class]) < 1e-6, (
            sound_class,
            single_class_psds[sound_class],
            best_psds_values[sound_class],
        )


def test_median_filter_oracle_tuning_psds_and_cbf():
    scores = io.read_sed_scores(package_dir / "tests" / "dcase2023task4a" / "baseline2_run1_scores")
    ground_truth = io.read_ground_truth_events(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "ground_truth.tsv"
    )
    ground_truth = {audio_id: ground_truth[audio_id] for audio_id in ground_truth if audio_id in scores}
    scores = {audio_id: scores[audio_id] for audio_id in ground_truth}
    audio_durations = io.read_audio_durations(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "audio_durations.tsv"
    )
    audio_durations = {audio_id: audio_durations[audio_id] for audio_id in scores}

    # tune hyper-parameters on subset of DESED's public eval
    predictors = median_filter.tune(
        scores=scores,
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        filter_lengths=np.linspace(0.0, 2.0, 11),
        selection_fn=median_filter.select_best_psds_and_cbf,
    )

    # psds predictor
    median_filter_psds, best_psds_values = predictors["psds"]
    # run inference on subset of DESED's public eval
    sed_scores = median_filter_psds.predict(scores)
    # evaluate PSDS performance
    psds, single_class_psds, *_ = intersection_based.psds(
        scores=sed_scores,
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        dtc_threshold=0.7,
        gtc_threshold=0.7,
        alpha_ct=0.0,
        alpha_st=1.0,
        unit_of_time="hour",
        max_efpr=100.0,
    )
    for sound_class in single_class_psds:
        assert abs(single_class_psds[sound_class] - best_psds_values[sound_class]) < 1e-6, (
            sound_class,
            single_class_psds[sound_class],
            best_psds_values[sound_class],
        )

    # cbf predictor
    median_filter_cbf, best_cbf_values = predictors["cbf"]

    # the following two blocks should give the same results:
    # 1) soft score inference and thresholding in collar_based.fscore
    sed_scores = median_filter_cbf.predict(scores)
    f, *_ = collar_based.fscore(
        scores=sed_scores,
        ground_truth=ground_truth,
        threshold=median_filter_cbf.detection_threshold,
        onset_collar=0.2,
        offset_collar=0.2,
        offset_collar_rate=0.2,
    )
    for sound_class in median_filter_cbf.sound_classes:
        assert abs(f[sound_class] - best_cbf_values[sound_class]) < 1e-6, (
            sound_class,
            f[sound_class],
            best_cbf_values[sound_class],
        )

    # 2) binary score inference (detection) and simple 0.5 threshold in collar_based.fscore
    sed_scores = median_filter_cbf.detect(scores, return_sed_scores=True)
    f, *_ = collar_based.fscore(
        scores=sed_scores,
        ground_truth=ground_truth,
        threshold=0.5,  # can be any number between 0 and 1 here as the sed_scores obtained from detection are binary
        onset_collar=0.2,
        offset_collar=0.2,
        offset_collar_rate=0.2,
    )
    for sound_class in median_filter_cbf.sound_classes:
        assert abs(f[sound_class] - best_cbf_values[sound_class]) < 1e-6, (
            sound_class,
            f[sound_class],
            best_cbf_values[sound_class],
        )


def test_median_filter_cross_validation_psds():
    # load predictions, ground_truth, and audio_durations
    scores = io.read_sed_scores(package_dir / "tests" / "dcase2023task4a" / "baseline2_run1_scores")
    ground_truth = io.read_ground_truth_events(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "ground_truth.tsv"
    )
    ground_truth = {audio_id: ground_truth[audio_id] for audio_id in ground_truth if audio_id in scores}
    scores = {audio_id: scores[audio_id] for audio_id in ground_truth}
    audio_durations = io.read_audio_durations(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "audio_durations.tsv"
    )
    audio_durations = {audio_id: audio_durations[audio_id] for audio_id in scores}

    # 5-fold cross validation split
    n_folds = 5
    all_keys = scores.keys()
    shuffled_keys = sorted(all_keys)
    np.random.RandomState(0).shuffle(shuffled_keys)
    split_idx = np.linspace(0, len(shuffled_keys), n_folds + 1).astype(int).tolist()
    folds = [set(shuffled_keys[split_idx[i] : split_idx[i + 1]]) for i in range(n_folds)]

    # cross validation
    csebbs_predictors, sed_scores, detections = median_filter.cross_validation(
        scores=scores,
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        folds=folds,
        filter_lengths=np.linspace(0.0, 2.0, 11),
        selection_fn=median_filter.select_best_psds,
        return_sed_scores=True,
    )

    # evaluate PSDS performance
    psds, single_class_psds, *_ = intersection_based.psds(
        scores=sed_scores,
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        dtc_threshold=0.7,
        gtc_threshold=0.7,
        alpha_ct=0.0,
        alpha_st=1.0,
        unit_of_time="hour",
        max_efpr=100.0,
    )
    # comment assert as evaluating reduced set (20 files) only
    # assert abs(psds - 0.5839223860184434) < 1e-6, psds  # result from paper


def test_median_filter_cross_validation_psds_cbf():
    # load predictions, ground_truth, and audio_durations
    scores = io.read_sed_scores(package_dir / "tests" / "dcase2023task4a" / "baseline2_run1_scores")
    ground_truth = io.read_ground_truth_events(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "ground_truth.tsv"
    )
    ground_truth = {audio_id: ground_truth[audio_id] for audio_id in ground_truth if audio_id in scores}
    scores = {audio_id: scores[audio_id] for audio_id in ground_truth}
    audio_durations = io.read_audio_durations(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "audio_durations.tsv"
    )
    audio_durations = {audio_id: audio_durations[audio_id] for audio_id in scores}

    # 5-fold cross validation split
    n_folds = 5
    all_keys = scores.keys()
    shuffled_keys = sorted(all_keys)
    np.random.RandomState(0).shuffle(shuffled_keys)
    split_idx = np.linspace(0, len(shuffled_keys), n_folds + 1).astype(int).tolist()
    folds = [set(shuffled_keys[split_idx[i] : split_idx[i + 1]]) for i in range(n_folds)]

    # cross validation
    median_filters, sed_scores, detections = median_filter.cross_validation(
        scores=scores,
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        folds=folds,
        filter_lengths=np.linspace(0.0, 2.0, 11),
        selection_fn=median_filter.select_best_psds_and_cbf,
        return_sed_scores=True,
    )

    # evaluate PSDS performance
    psds, single_class_psds, *_ = intersection_based.psds(
        scores=sed_scores["psds"],
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        dtc_threshold=0.7,
        gtc_threshold=0.7,
        alpha_ct=0.0,
        alpha_st=1.0,
        unit_of_time="hour",
        max_efpr=100.0,
    )
    # comment assert as evaluating reduced set (20 files) only
    # assert abs(psds - 0.5839223860184434) < 1e-6, psds  # result from paper

    # evaluate collar-based F1-score performance
    f, *_ = collar_based.fscore(
        scores=detections["cbf"],
        ground_truth=ground_truth,
        threshold=0.5,  # can be any number between 0 and 1 here as the sed_scores obtained from detection are binary
        onset_collar=0.2,
        offset_collar=0.2,
        offset_collar_rate=0.2,
    )
    # comment assert as evaluating reduced set (20 files) only
    # assert abs(f['macro_average'] - 0.6504733960881937) < 1e-6, f['macro_average']  # result from paper



# Copyright (C) 2024 Mitsubishi Electric Research Laboratories (MERL)
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import numpy as np

from sebbs.change_detection import change_detection


def test_vs_discrete_filter():
    N = 500
    fil_len = 1.0
    y = np.random.RandomState(0).rand(N, 2)
    timestamps = np.linspace(0.0, 10.0, N + 1)
    changes = change_detection(y, timestamps, fil_len)

    step_filter = np.ones(N // 10)
    step_filter[: N // 20] = -1
    for i, (segment_bounds, *_) in enumerate(changes):
        onsets = segment_bounds[1:-2:2]
        offsets = segment_bounds[2:-1:2]
        deltas_ref = np.correlate(np.concatenate((np.zeros(N // 20), y[:, i], np.zeros(N // 20))), step_filter)
        deltas_ref = np.concatenate(([-np.inf], deltas_ref, [-np.inf]))
        onsets_ref = timestamps[
            np.argwhere((deltas_ref[1:-2] > deltas_ref[:-3]) * (deltas_ref[1:-2] > deltas_ref[2:-1])).flatten()
        ]
        offsets_ref = timestamps[
            np.argwhere((deltas_ref[2:-1] < deltas_ref[1:-2]) * (deltas_ref[2:-1] < deltas_ref[3:])).flatten() + 1
        ]
        if len(onsets_ref) > len(offsets_ref):
            offsets_ref = np.concatenate((offsets_ref, [timestamps[-1]]))
        assert (np.abs(onsets - onsets_ref) < 1e-6).all()
        assert (np.abs(offsets - offsets_ref) < 1e-6).all()



# Copyright (C) 2024 Mitsubishi Electric Research Laboratories (MERL)
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from pathlib import Path
from typing import Callable, Iterable, Tuple, Union

import numpy as np
from sed_scores_eval import collar_based, intersection_based, io
from sed_scores_eval.base_modules.detection import scores_to_event_list
from sed_scores_eval.base_modules.postprocessing import medfilt
from sed_scores_eval.base_modules.scores import validate_score_dataframe

from sebbs.utils import sed_scores_from_detections


class MedianFilter:
    def __init__(
        self,
        filter_length: Union[float, dict],
        detection_threshold: Union[float, dict, None] = None,
        sound_classes: Union[list, None] = None,
    ):
        """wrapper around sed_scores_eval's medfilt

        Args:
            filter_length (float|dict): (class-wise) median filter length(s).
            detection_threshold (float|dict|None): (class-wise) detection/decision thresholds.
            sound_classes (list of string|None): sound classes
        """
        self.filter_length = filter_length
        self.detection_threshold = detection_threshold
        self.sound_classes = sound_classes

    def copy(self):
        """get copy of the class instance (CSEBBsPredictor)"""
        return MedianFilter(
            {**self.filter_length} if isinstance(self.filter_length, dict) else self.filter_length,
            sound_classes=self.sound_classes,
        )

    def predict(
        self,
        scores: Union[str, dict],
        audio_ids: Union[Iterable, None] = None,
    ) -> dict:
        """run median filtering

        Args:
            scores (dict of pandas.DataFrame): SED posterior score data frames
                as used in sed_scores_eval (https://github.com/fgnt/sed_scores_eval).
            audio_ids (iterable): audio ids that prediction should be restricted to.

        Returns (dict of pd.DataFrame):
            median filtered SED posterior score data frames for each audio_id

        """
        if isinstance(scores, (str, Path)):
            scores = io.read_sed_scores(scores)
        if audio_ids is None:
            audio_ids = scores.keys()
        scores = {audio_id: scores[audio_id] for audio_id in audio_ids}
        _, self.sound_classes = validate_score_dataframe(scores[list(audio_ids)[0]], event_classes=self.sound_classes)

        if isinstance(self.filter_length, dict):
            filter_len = np.array([self.filter_length[sound_class] for sound_class in self.sound_classes])
        else:
            assert np.isscalar(self.filter_length), self.filter_length
            filter_len = self.filter_length
        return medfilt(scores, filter_len)

    def detect(
        self,
        scores: Union[str, dict],
        detection_threshold: Union[float, dict, None] = None,
        audio_ids: Union[Iterable, None] = None,
        return_sed_scores: bool = False,
    ) -> dict:
        """perform median filtering and derive detections/decisions by employing detection_threshold

        Args:
            scores (str | dict of pandas.DataFrame): (path to directory with)
                SED posterior score data frames as used in sed_scores_eval
                (https://github.com/fgnt/sed_scores_eval).
            detection_threshold (float|dict|None): (class-wise) detection/decision thresholds.
                If None, self.detection_threshold is used.
            audio_ids (iterable): audio ids that prediction should be restricted to.
            return_sed_scores (bool): whether to return outputs as SED scores
                data frame format for direct use with sed_scores_eval-based evaluation.
                In that case, sed_scores are binary due to detection thresholding.

        Returns (dict of list):
            a list of detected sound events (onset, offset, class) (which may have
            been converted to a SED scores data frame) for each audio_id

        """
        if detection_threshold is None:
            detection_threshold = self.detection_threshold
        assert detection_threshold is not None, "A detection threshold has to be provided to run detection."

        median_filtered_scores = self.predict(scores, audio_ids)

        return self.detection_thresholding(
            median_filtered_scores,
            detection_threshold,
            return_sed_scores=return_sed_scores,
        )

    def detection_thresholding(
        self,
        median_filtered_scores: dict,
        detection_threshold: Union[float, dict, None] = None,
        return_sed_scores: bool = False,
    ) -> dict:
        """

        Args:
            median_filtered_scores (dict of pd.DataFrame): median filtered SED posterior scores for each audio_id
            detection_threshold (float|dict|None): (class-wise) detection/decision thresholds.
                If None, self.detection_threshold is used.:
            return_sed_scores (bool): whether to return outputs as SED scores
                data frame format for direct use with sed_scores_eval-based evaluation.
                In that case, sed_scores are binary due to detection thresholding.

        Returns (dict of list):
            a list of detected sound events (onset, offset, class) (which may have
            been converted to a SED scores data frame) for each audio_id

        """
        if detection_threshold is None:
            detection_threshold = self.detection_threshold
        assert detection_threshold is not None, "A detection threshold has to be provided to run detection."
        detections = scores_to_event_list(median_filtered_scores, detection_threshold, self.sound_classes)
        if return_sed_scores:
            detections = sed_scores_from_detections(
                detections,
                sound_classes=self.sound_classes,
            )
        return detections


def tune(
    scores: Union[str, dict],
    ground_truth: Union[str, dict],
    audio_durations: Union[str, dict],
    *,
    filter_lengths: Iterable = np.linspace(0.0, 2.0, 11),
    selection_fn: Callable,
    folds: Union[Iterable, None] = None,
    **selection_kwargs,
) -> Union[Tuple[MedianFilter, dict], list]:
    """perform grid search over hyper-parameters

    Args:
        scores (str | dict of pandas.DataFrame): (path to directory with)
            SED posterior score data frames as used in sed_scores_eval
            (https://github.com/fgnt/sed_scores_eval).
        ground_truth (str | dict of list): path to/dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio as used in
            sed_scores_eval.
        audio_durations (str | dict): (path to) durations of each audio file.
        filter_lengths (iterable): candidates for median filter lengths.
        selection_fn (callable): function that takes list of
            (CSEBBsPredictor, Sebb predictions) tuples together with
            ground truth, audio durations, the audio ids to be considered and
            further optional arguments and selects best CSEBBsPredictor parameters.
        folds (list|None): optional list of audio_id sets for which best parameters
            should be selected separately in which case a list of
            CSEBBsPredictors is returned. If None, best parameters are selected
            based on the whole set and a single CSEBBsPredictor is returned.
        **selection_kwargs: any key-word arguments that should be forwarded to selection_fn.

    Returns: (list of) return(s) from selection_fn

    """
    if isinstance(scores, (str, Path)):
        scores = io.read_sed_scores(scores)
    if isinstance(ground_truth, (str, Path)):
        ground_truth = io.read_ground_truth_events(ground_truth)
    if isinstance(audio_durations, (str, Path)):
        audio_durations = io.read_audio_durations(audio_durations)
    median_filtered_scores = []
    for filt_len in filter_lengths:
        median_filtered_scores.append((filt_len, medfilt(scores, filt_len)))
    if folds is None:
        return selection_fn(median_filtered_scores, ground_truth, audio_durations, **selection_kwargs)
    return [
        selection_fn(median_filtered_scores, ground_truth, audio_durations, audio_ids=fold, **selection_kwargs)
        for fold in folds
    ]


def cross_validation(
    scores: Union[str, dict],
    ground_truth: Union[str, dict],
    audio_durations: Union[str, dict],
    folds: Iterable,
    *,
    filter_lengths: Iterable = np.linspace(0.0, 2.0, 11),
    selection_fn: Callable,
    return_sed_scores: bool = True,
    **selection_kwargs,
) -> Tuple[list, dict, dict]:
    """perform leave-one-out cross validation. I.e., tune hyper-parameters on
    all folds but one and write outputs for the left out fold.

    Args:
        scores (str | dict of pandas.DataFrame): (path to directory with)
            SED posterior score data frames as used in sed_scores_eval
            (https://github.com/fgnt/sed_scores_eval).
        ground_truth (str | dict of list): path to/dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio as used in
            sed_scores_eval.
        audio_durations (str | dict): (path to) durations of each audio file.
        folds (iterable): optional list of audio_id sets for which best parameters
            should be selected separately in which case a list of
            CSEBBsPredictors is returned. If None, best parameters are selected
            based on the whole set and a single CSEBBsPredictor is returned.
        filter_lengths (iterable): candidates for median filter lengths.
        selection_fn (callable): function that takes list of
            (CSEBBsPredictor, Sebb predictions) tuples together with
            ground truth, audio durations, the audio ids to be considered and
            further optional arguments and selects best CSEBBsPredictor parameters.
        return_sed_scores (bool): whether to return outputs in sed_scores_eval
            format for direct use with sed_scores_eval-based evaluation.
        **selection_kwargs: any key-word arguments that should be forwarded to selection_fn.

    Returns:
        csebbs_predictors (list): list of return(s) from selection_fn
        sebbs (dict): predicted SEBBs for the whole data set,
            where predictions for different folds used different tuned hyper-parameters
        detections (dict|None): Detected events for the whole data set,
            if detection thresholds are selected by selection_fn else None.

    """
    if isinstance(scores, (str, Path)):
        scores = io.read_sed_scores(scores)
    if isinstance(ground_truth, (str, Path)):
        ground_truth = io.read_ground_truth_events(ground_truth)
    if isinstance(audio_durations, (str, Path)):
        audio_durations = io.read_audio_durations(audio_durations)
    test_ids = [set(fold) for fold in folds]
    all_ids = set.union(*test_ids)
    val_ids = [(all_ids - fold) for fold in test_ids]

    median_filters = tune(
        scores,
        ground_truth,
        audio_durations,
        folds=val_ids,
        filter_lengths=filter_lengths,
        selection_fn=selection_fn,
        **selection_kwargs,
    )
    median_filtered_scores = {}
    detections = {}
    for i, predictors in enumerate(median_filters):
        if not isinstance(predictors, dict):
            predictors = {"": predictors}
        for key, predictor in predictors.items():
            if isinstance(predictor, tuple):
                predictor = predictor[0]
            assert isinstance(predictor, MedianFilter), "selection_fn doesn't seem to have returned a CSEBBsPredictor."
            median_filtered_scores_i = predictor.predict(
                scores,
                audio_ids=test_ids[i],
            )
            if key not in median_filtered_scores:
                median_filtered_scores[key] = {}
                detections[key] = {}
            median_filtered_scores[key].update(median_filtered_scores_i)
            if predictor.detection_threshold is not None:
                detections_i = predictor.detection_thresholding(
                    median_filtered_scores_i,
                    return_sed_scores=return_sed_scores,
                )
                detections[key].update(detections_i)
    if len(median_filtered_scores) == 1 and "" in median_filtered_scores:
        median_filtered_scores = median_filtered_scores[""]
        detections = detections[""]
    return median_filters, median_filtered_scores, detections


def select_best_psds(
    median_filtered_scores: list,
    ground_truth: dict,
    audio_durations: dict,
    audio_ids: Union[Iterable, None] = None,
    dtc_threshold: float = 0.7,
    gtc_threshold: float = 0.7,
    cttc_threshold: Union[float, None] = None,
    alpha_ct: float = 0.0,
    unit_of_time: str = "hour",
    max_efpr: float = 100.0,
    classwise: bool = True,
) -> Tuple[MedianFilter, dict]:
    """select parameters which give highest PSDS

    Args:
        median_filtered_scores (list): list of (filter length, filtered scores) provided from tune function above
        ground_truth (dict of list): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio as used in
            sed_scores_eval.
        audio_durations: The duration of each audio file.
        audio_ids (iterable|None): audio ids that parameter selection should be restricted to.
        dtc_threshold (float): detection tolerance criterion threshold of PSDS
        gtc_threshold (float): ground truth intersection criterion threshold of PSDS
        cttc_threshold (float|None): cross trigger tolerance criterion threshold of PSDS
        alpha_ct (float): cross trigger penalization weight of PSDS.
        unit_of_time (str): time unit for FPR computation in PSDS
        max_efpr (float): maximum false positives per time unit to be considered in PSD-ROC.
        classwise (bool): whether to use different parameters for each sound_class or not.

    Returns:
        median_filter: MedianFilter with best hyper-parameters w.r.t. PSDS
        best_values: class-wise PSDS values with best hyper-parameters

    """
    if audio_ids is not None:
        median_filtered_scores = [
            (filter_len, {audio_id: scores_i[audio_id] for audio_id in audio_ids})
            for filter_len, scores_i in median_filtered_scores
        ]
        ground_truth = {audio_id: ground_truth[audio_id] for audio_id in audio_ids}
        audio_durations = {audio_id: audio_durations[audio_id] for audio_id in audio_ids}

    best_filter_length = {}
    best_values = {}
    for filter_len, scores_i in median_filtered_scores:
        single_class_psds = intersection_based.psds(
            scores=scores_i,
            ground_truth=ground_truth,
            audio_durations=audio_durations,
            dtc_threshold=dtc_threshold,
            gtc_threshold=gtc_threshold,
            cttc_threshold=cttc_threshold,
            alpha_ct=alpha_ct,
            unit_of_time=unit_of_time,
            max_efpr=max_efpr,
        )[1]
        mean = np.mean(list(single_class_psds.values()))
        for sound_class in single_class_psds:
            value = single_class_psds[sound_class] if classwise else mean
            if (sound_class not in best_values) or (value > best_values[sound_class]):
                # for this sound class the performance achieved with current params is better than with previous params
                best_values[sound_class] = value
                best_filter_length[sound_class] = filter_len

    median_filter = MedianFilter(
        filter_length=best_filter_length,
    )
    return median_filter, best_values


def select_best_cbf(
    median_filtered_scores: list,
    ground_truth: dict,
    audio_durations: dict,
    audio_ids: Union[Iterable, None] = None,
    onset_collar: float = 0.2,
    offset_collar: float = 0.2,
    offset_collar_rate: float = 0.2,
    classwise: bool = True,
) -> Tuple[MedianFilter, dict]:
    """select parameters which give highest collar-based F1-score

    Args:
        median_filtered_scores (list): list of (filter length, filtered scores) provided from tune function above
        ground_truth (dict of list): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio as used in
            sed_scores_eval.
        audio_durations: The duration of each audio file.
        audio_ids (iterable|None): audio ids that parameter selection should be restricted to.
        onset_collar (float): onset collar in seconds to be used in collar-based evaluation.
        offset_collar (float): min offset collar in seconds to be used in collar-based evaluation.
        offset_collar_rate (float): min offset collar as a rate of the ground truth event length
            to be used in collar-based evaluation. Actual offset collar given as
            offset_collar_for_gt_event = max(
                offset_collar, offset_collar_rate*length_of_gt_event_in_seconds
            )
        classwise (bool): whether to use different parameters for each sound_class or not.

    Returns:
        median_filter: MedianFilter with best hyper-parameters w.r.t. collar-based F1-score
        best_values: class-wise collar-based F1-score values with best hyper-params

    """
    if audio_ids is not None:
        median_filtered_scores = [
            (filter_len, {audio_id: scores_i[audio_id] for audio_id in audio_ids})
            for filter_len, scores_i in median_filtered_scores
        ]
        ground_truth = {audio_id: ground_truth[audio_id] for audio_id in audio_ids}
        audio_durations = {audio_id: audio_durations[audio_id] for audio_id in audio_ids}

    best_filter_length = {}
    best_detection_threshold = {}
    best_values = {}
    for filter_len, scores_i in median_filtered_scores:
        f, _, _, thresholds_cbf, _ = collar_based.best_fscore(
            scores=scores_i,
            ground_truth=ground_truth,
            onset_collar=onset_collar,
            offset_collar=offset_collar,
            offset_collar_rate=offset_collar_rate,
        )
        for sound_class in thresholds_cbf:
            value = f[sound_class] if classwise else f["macro_average"]
            if (sound_class not in best_values) or (value > best_values[sound_class]):
                # for this sound class the performance achieved with current params is better than with previous params
                best_values[sound_class] = value
                best_filter_length[sound_class] = filter_len
                best_detection_threshold[sound_class] = thresholds_cbf[sound_class]

    median_filter = MedianFilter(
        filter_length=best_filter_length,
        detection_threshold=best_detection_threshold,
    )
    return median_filter, best_values


def select_best_psds_and_cbf(
    median_filtered_scores: list,
    ground_truth: dict,
    audio_durations: dict,
    audio_ids: Union[Iterable, None] = None,
    dtc_threshold: float = 0.7,
    gtc_threshold: float = 0.7,
    cttc_threshold: Union[float, None] = None,
    alpha_ct: float = 0.0,
    unit_of_time: str = "hour",
    max_efpr: float = 100.0,
    onset_collar: float = 0.2,
    offset_collar: float = 0.2,
    offset_collar_rate: float = 0.2,
    classwise: bool = True,
) -> dict:
    """Select both, best PSDS and collar-based parameters

    Args:
        median_filtered_scores (list): list of (filter length, filtered scores) provided from tune function above
        ground_truth (dict of list): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio as used in
            sed_scores_eval.
        audio_durations: The duration of each audio file.
        audio_ids (iterable|None): audio ids that parameter selection should be restricted to.
        dtc_threshold (float): detection tolerance criterion threshold of PSDS
        gtc_threshold (float): ground truth intersection criterion threshold of PSDS
        cttc_threshold (float|None): cross trigger tolerance criterion threshold of PSDS
        alpha_ct (float): cross trigger penalization weight of PSDS.
        unit_of_time (str): time unit for FPR computation in PSDS
        max_efpr (float): maximum false positives per time unit to be considered in PSD-ROC.
        onset_collar (float): onset collar in seconds to be used in collar-based evaluation.
        offset_collar (float): min offset collar in seconds to be used in collar-based evaluation.
        offset_collar_rate (float): min offset collar as a rate of the ground truth event length
            to be used in collar-based evaluation. Actual offset collar given as
            offset_collar_for_gt_event = max(
                offset_collar, offset_collar_rate*length_of_gt_event_in_seconds
            )
        classwise (bool): whether to use different parameters for each sound_class or not.

    Returns: dict of (tuned MedianFilter, class-wise performance dict) tuples
        for both PSDS and collar-based F1-score metrics

    """
    return {
        "psds": select_best_psds(
            median_filtered_scores,
            ground_truth,
            audio_durations,
            audio_ids=audio_ids,
            dtc_threshold=dtc_threshold,
            gtc_threshold=gtc_threshold,
            cttc_threshold=cttc_threshold,
            alpha_ct=alpha_ct,
            unit_of_time=unit_of_time,
            max_efpr=max_efpr,
            classwise=classwise,
        ),
        "cbf": select_best_cbf(
            median_filtered_scores,
            ground_truth,
            audio_durations,
            audio_ids=audio_ids,
            onset_collar=onset_collar,
            offset_collar=offset_collar,
            offset_collar_rate=offset_collar_rate,
            classwise=classwise,
        ),
    }



# Copyright (C) 2024 Mitsubishi Electric Research Laboratories (MERL)
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import itertools
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Tuple, Union

import numpy as np
from sed_scores_eval import collar_based, intersection_based, io
from sed_scores_eval.base_modules.scores import validate_score_dataframe
from tqdm import tqdm

from sebbs.change_detection import change_detection
from sebbs.utils import sed_scores_from_detections, sed_scores_from_sebbs

class CSEBBsPredictor:
    """change-point based predictor of Sound Event Bounding Boxes (cSEBBs).

    [1] J.Ebbers, F.Germain, G.Wichern and J.Le Roux, "Sound Event Bounding Boxes",
        Accepted for publication at Interspeech, 2024

    """

    def __init__(
        self,
        step_filter_length: Union[float, dict] = 0.5,
        merge_threshold_abs: Union[float, dict] = 1.0,
        merge_threshold_rel: Union[float, dict] = 2.0,
        detection_threshold: Union[float, dict, None] = None,
        sound_classes: Union[list, None] = None,
    ):
        """

        Args:
            step_filter_length (float|dict): (class-wise) step filter length(s)
                for change point detection.
            merge_threshold_abs (float|dict): (class-wise) absolute threshold(s) for segment merging.
                If the absolute difference between a min value in a gap segment
                and the max value in the neighbouring event candidates is
                smaller than this threshold, the three segments (two events with
                a gap in between) may be merged (if rel_threshold is fulfilled also)
            merge_threshold_rel (float|dict): (class-wise) relative threshold(s) for segment merging.
                If the relative difference between a min value in a gap segment
                and the max value in the neighbouring event candidates (max value/min_value)
                is smaller than this threshold, the three segments (two events with
                a gap in between) may be merged (if abs_threshold is fulfilled also)
            detection_threshold (float|dict|None): (class-wise) detection/decision thresholds.
            sound_classes (list of string|None): sound classes
        """
        self.step_filter_length = step_filter_length
        self.merge_threshold_abs = merge_threshold_abs
        self.merge_threshold_rel = merge_threshold_rel
        self.detection_threshold = detection_threshold
        self.sound_classes = sound_classes

    def to_dict(self) -> Dict[str, Any]:
        return {
            "step_filter_length": self.step_filter_length,
            "merge_threshold_abs": self.merge_threshold_abs,
            "merge_threshold_rel": self.merge_threshold_rel,
            "detection_threshold": self.detection_threshold,
            "sound_classes": self.sound_classes,
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "CSEBBsPredictor":
        return cls(
            step_filter_length=d["step_filter_length"],
            merge_threshold_abs=d["merge_threshold_abs"],
            merge_threshold_rel=d["merge_threshold_rel"],
            detection_threshold=d["detection_threshold"],
            sound_classes=d["sound_classes"],
        )

    def copy(self):
        """get copy of the class instance (CSEBBsPredictor)"""
        return CSEBBsPredictor(
            {**self.step_filter_length} if isinstance(self.step_filter_length, dict) else self.step_filter_length,
            {**self.merge_threshold_abs} if isinstance(self.merge_threshold_abs, dict) else self.merge_threshold_abs,
            {**self.merge_threshold_rel} if isinstance(self.merge_threshold_rel, dict) else self.merge_threshold_rel,
            sound_classes=self.sound_classes,
        )

    def predict(
        self,
        scores: Union[str, dict],
        audio_ids: Union[Iterable, None] = None,
        sound_classes: Union[list, None] = None,
        return_sed_scores: bool = False,
    ) -> dict:
        """run post-processing on scores to predict sebbs

        Args:
            scores (str | dict of pandas.DataFrame): (path to directory with)
                SED posterior score data frames as used in sed_scores_eval
                (https://github.com/fgnt/sed_scores_eval).
            audio_ids (Iterable|None): audio ids that prediction should be restricted to.
            return_sed_scores (bool): whether to return outputs as SED scores
                data frame format for direct use with sed_scores_eval-based evaluation.

        Returns (dict of list|pd.DataFrame):
            a list of SEBBs (onset, offset, class) (which may have
            been converted to a SED scores data frame) for each audio_id

        """
        if isinstance(scores, (str, Path)):
            scores = io.read_sed_scores(scores)
        if audio_ids is None:
            audio_ids = scores.keys()
        scores = {audio_id: scores[audio_id] for audio_id in audio_ids}
        _, sound_classes = validate_score_dataframe(scores[list(audio_ids)[0]], event_classes=sound_classes or self.sound_classes)
        change_detection = self._run_change_detection(scores, sound_classes)
        sebbs = self._get_sebbs_from_change_detection(change_detection, sound_classes)
        if return_sed_scores:
            return sed_scores_from_sebbs(sebbs, sound_classes)
        return sebbs

    def detect(
        self,
        scores: Union[str, dict],
        detection_threshold: Union[float, dict, None] = None,
        audio_ids: Union[Iterable, None] = None,
        return_sed_scores: bool = False,
    ) -> dict:
        """predict SEBBs and derive detections/decisions by employing detection_threshold

        Args:
            scores (str | dict of pandas.DataFrame): (path to directory with)
                SED posterior score data frames as used in sed_scores_eval
                (https://github.com/fgnt/sed_scores_eval).
            detection_threshold (float|dict|None): (class-wise) detection/decision thresholds.
                If None, self.detection_threshold is used.
            audio_ids (iterable|None): audio ids that prediction should be restricted to.
            return_sed_scores (bool): whether to return outputs as SED scores
                data frame format for direct use with sed_scores_eval-based evaluation.
                In that case, sed_scores are binary due to detection thresholding.

        Returns (dict of list):
            a list of detected sound events (onset, offset, class) (which may have
            been converted to a SED scores data frame) for each audio_id

        """
        if detection_threshold is None:
            detection_threshold = self.detection_threshold
        assert detection_threshold is not None, "A detection threshold has to be provided to run detection."

        sebbs = self.predict(scores, audio_ids)

        return self.detection_thresholding(
            sebbs,
            detection_threshold,
            return_sed_scores=return_sed_scores,
        )

    def detection_thresholding(
        self,
        sebbs: dict,
        detection_threshold: Union[float, dict, None] = None,
        return_sed_scores: bool = False,
    ) -> dict:
        """

        Args:
            sebbs (dict of list): a list of SEBBs (onset, offset, class, confidence) for each audio_id
            detection_threshold (float|dict|None): (class-wise) detection/decision thresholds.
                If None, self.detection_threshold is used.
            return_sed_scores (bool): whether to return outputs as SED scores
                data frame format for direct use with sed_scores_eval-based evaluation.
                In that case, sed_scores are binary due to detection thresholding.

        Returns (dict of list):
            a list of detected sound events (onset, offset, class) (which may have
            been converted to a SED scores data frame) for each audio_id

        """
        if detection_threshold is None:
            detection_threshold = self.detection_threshold
        assert detection_threshold is not None, "A detection threshold has to be provided to run detection."
        if not isinstance(detection_threshold, dict):
            assert np.isscalar(detection_threshold), detection_threshold
            detection_threshold = {sound_class: detection_threshold for sound_class in self.sound_classes}
        detections = {
            audio_id: [sebb[:3] for sebb in sebbs_i if sebb[3] > detection_threshold[sebb[2]]]
            for audio_id, sebbs_i in sebbs.items()
        }
        if return_sed_scores:
            detections = sed_scores_from_detections(
                detections,
                sound_classes=self.sound_classes,
            )
        return detections

    def _run_change_detection(self, scores, sound_classes: Union[list, None] = None,):
        """detect candidate segment boundaries/change points"""
        sound_classes = sound_classes or self.sound_classes
        if isinstance(self.step_filter_length, dict):
            step_filter_length = np.array([self.step_filter_length[sound_class] for sound_class in sound_classes])
        else:
            step_filter_length = self.step_filter_length
        change_detection_out = {}
        for key, scores_df in scores.items():
            timestamps, sound_classes = validate_score_dataframe(scores_df, event_classes=sound_classes)
            scores_arr = scores_df[sound_classes].to_numpy()
            change_detection_out[key] = change_detection(scores_arr, timestamps, step_filter_length)
        for key, det in change_detection_out.items():
            change_detection_out[key] = {
                sound_class: change_detection_out[key][c] for c, sound_class in enumerate(sound_classes)
            }
        return change_detection_out

    def _get_sebbs_from_change_detection(self, change_detection, sound_classes: Union[list, None] = None,):
        """perform merging of segments and infer SEBBs"""
        sound_classes = sound_classes or self.sound_classes
        sebbs = {}
        for audio_id in change_detection.keys():
            onsets = []
            offsets = []
            confidences = []
            class_labels = []
            for k, sound_class in enumerate(sound_classes):
                seg_bounds, mean_scores_k, min_scores_k, max_scores_k = change_detection[audio_id][sound_class]
                abs_thres = (
                    self.merge_threshold_abs[sound_class]
                    if isinstance(self.merge_threshold_abs, dict)
                    else self.merge_threshold_abs
                )
                rel_thres = (
                    self.merge_threshold_rel[sound_class]
                    if isinstance(self.merge_threshold_rel, dict)
                    else self.merge_threshold_rel
                )
                onsets_c, offsets_c, confidences_c = _merge_segments(
                    seg_bounds,
                    mean_scores_k,
                    min_scores_k,
                    max_scores_k,
                    threshold_abs=abs_thres,
                    threshold_rel=rel_thres,
                )
                onsets.extend(onsets_c)
                offsets.extend(offsets_c)
                class_labels.extend(len(onsets_c) * [sound_class])
                confidences.extend(confidences_c)
            sebbs[audio_id] = [
                (onset, offset, class_label, confidence)
                for onset, offset, class_label, confidence in zip(
                    onsets,
                    offsets,
                    class_labels,
                    confidences,
                )
            ]
        return sebbs


def _merge_segments(
    seg_bounds,
    mean_scores,
    min_scores,
    max_scores,
    threshold_abs=0.5,
    threshold_rel=2.0,
):
    """perform merging of segments/candidate events when scores differ less than an absolute and/or relative threshold.

    Args:
        seg_bounds: candidate segment boundaries/change points.
            seg_bounds[0] equals audio onset 0., seg_bounds[1:-2:2] are event
            onset candidates, seg_bounds[2:-1:2] are event offset candidates,
            and seg_bounds[-1] equals audio length.
        mean_scores (1d np.ndarray): candidate segments' mean scores
        min_scores (1d np.ndarray): candidate segments' min scores
        max_scores (1d np.ndarray): candidate segments' max scores
        threshold_abs (float): the absolute threshold for segment merging.
            If the absolute difference between a min value in a gap segment
            and the max value in the neighbouring event candidates is
            smaller than this threshold, the three segments (two events with
            a gap in between) may be merged (if threshold_rel is fulfilled also).
        threshold_rel (float): the relative threshold for segment merging.
            If the relative difference between a min value in a gap segment
            and the max value in the neighbouring event candidates (max value/min_value)
            is smaller than this threshold, the three segments (two events with
            a gap in between) may be merged (if threshold_abs is fulfilled also).

    Returns:

    """
    if len(seg_bounds) >= 6:
        event_scores = max_scores[1:-1:2] + 1e-3
        gap_scores = min_scores[2:-2:2] + 1e-3
        gap_idx = np.argwhere(
            (
                ((event_scores[:-1] / gap_scores) < threshold_rel)
                * ((event_scores[1:] / gap_scores) < threshold_rel)
                * ((event_scores[:-1] - gap_scores) < threshold_abs)
                * ((event_scores[1:] - gap_scores) < threshold_abs)
            )
            < 0.5
        ).flatten()

        seg_idx = np.stack((2 + 2 * gap_idx, 2 + 2 * gap_idx + 1), axis=1).flatten()
        seg_idx = np.concatenate(([0, 1], seg_idx, [len(seg_bounds) - 2, len(seg_bounds) - 1])).astype(int)
        seg_bounds, mean_scores, min_scores, max_scores = _merge_scores(
            seg_bounds,
            mean_scores,
            min_scores,
            max_scores,
            seg_idx,
        )

    onsets = seg_bounds[1:-2:2]
    offsets = seg_bounds[2:-1:2]
    mean_scores = mean_scores[1:-1:2]
    return onsets, offsets, mean_scores


def _merge_scores(seg_bounds, mean_scores, min_scores, max_scores, seg_idx):
    """compute new mean,min and max score for segments after merging.

    Args:
        seg_bounds (1d np.ndarray): old segment bounds
        mean_scores (1d np.ndarray): old segments' mean scores
        min_scores (1d np.ndarray): old segments' min scores
        max_scores (1d np.ndarray): old segments' max scores
        seg_idx: indices of remaining segment boundaries

    Returns:
        seg_bounds (1d np.ndarray): new segment bounds
        mean_scores (1d np.ndarray): new segments' mean scores
        min_scores (1d np.ndarray): new segments' min scores
        max_scores (1d np.ndarray): new segments' max scores

    """
    seglens = seg_bounds[1:] - seg_bounds[:-1]
    mean_scores_new = []
    min_scores_new = []
    max_scores_new = []
    for j in range(0, len(seg_idx) - 1):
        if seglens[seg_idx[j] : seg_idx[j + 1]].sum() == 0.0:
            mean_scores_new.append(0.0)
            min_scores_new.append(0.0)
            max_scores_new.append(0.0)
        else:
            mean_scores_new.append(
                (seglens[seg_idx[j] : seg_idx[j + 1]] * mean_scores[seg_idx[j] : seg_idx[j + 1]]).sum()
                / (seglens[seg_idx[j] : seg_idx[j + 1]].sum() + 1e-6)
            )
            min_scores_new.append(min(min_scores[seg_idx[j] : seg_idx[j + 1]]))
            max_scores_new.append(max(max_scores[seg_idx[j] : seg_idx[j + 1]]))
    seg_bounds = seg_bounds[seg_idx]
    mean_scores = np.array(mean_scores_new)
    min_scores = np.array(min_scores_new)
    max_scores = np.array(max_scores_new)
    return seg_bounds, mean_scores, min_scores, max_scores


def _run(
    scores: dict,
    step_filter_lengths: List[float],
    merge_thresholds_abs: Iterable = (0.15, 0.2, 0.3),
    merge_thresholds_rel: Iterable = (1.5, 2.0, 3.0),
    either_abs_or_rel_threshold: bool = True,
):
    csebbs = []
    for step_filt_len in step_filter_lengths:
        csebbs_predictor = CSEBBsPredictor(
            step_filter_length=step_filt_len,
        )
        change_detection = csebbs_predictor._run_change_detection(scores)
        if either_abs_or_rel_threshold:
            it = [(thres_abs, np.inf) for thres_abs in merge_thresholds_abs] + [
                (np.inf, thres_rel) for thres_rel in merge_thresholds_rel
            ]
        else:
            it = itertools.product(merge_thresholds_abs, merge_thresholds_rel)
        for abs_thres, rel_thres in tqdm(it, desc="Get sebbs", total=len(it)):
            pred = csebbs_predictor.copy()
            pred.merge_threshold_abs = abs_thres
            pred.merge_threshold_rel = rel_thres
            csebbs.append((pred, pred._get_sebbs_from_change_detection(change_detection)))
    return csebbs


def tune(
    scores: Union[str, dict],
    ground_truth: Union[str, dict],
    audio_durations: Union[str, dict],
    *,
    step_filter_lengths: Iterable = (0.32, 0.48, 0.64),
    merge_thresholds_abs: Iterable = (0.15, 0.2, 0.3),
    merge_thresholds_rel: Iterable = (1.5, 2.0, 3.0),
    selection_fn: Callable,
    folds: Union[Iterable, None] = None,
    either_abs_or_rel_threshold: bool = True,
    num_jobs: int = 1,
    **selection_kwargs,
) -> Union[Tuple[CSEBBsPredictor, dict], list]:
    """perform grid search over hyper-parameters

    Args:
        scores (str | dict of pandas.DataFrame): (path to directory with)
            SED posterior score data frames as used in sed_scores_eval
            (https://github.com/fgnt/sed_scores_eval).
        ground_truth (str | dict of list): path to/dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio as used in
            sed_scores_eval.
        audio_durations (str | dict): (path to) durations of each audio file.
        step_filter_lengths (iterable): candidates for step_filter_length.
        merge_thresholds_abs (iterable): candidates for merge_threshold_abs.
        merge_thresholds_rel (iterable): candidates for merge_threshold_abs.
        selection_fn (callable): function that takes list of
            (CSEBBsPredictor, Sebb predictions) tuples together with
            ground truth, audio durations, the audio ids to be considered and
            further optional arguments and selects best CSEBBsPredictor parameters.
        folds (iterable|None): optional list of audio_id sets for which best parameters
            should be selected separately in which case a list of
            CSEBBsPredictors is returned. If None, best parameters are selected
            based on the whole set and a single CSEBBsPredictor is returned.
        either_abs_or_rel_threshold (bool): if True, either merge_thresholds_rel
            or merge_thresholds_abs is used. Whereas if False, they may be used
            jointly (which, however, increases the number of parameter
            combinations that need to be evaluated during tuning).
        **selection_kwargs: any key-word arguments that should be forwarded to selection_fn.

    Returns: (list of) return(s) from selection_fn

    """
    if isinstance(scores, (str, Path)):
        scores = io.read_sed_scores(scores)
    if isinstance(ground_truth, (str, Path)):
        ground_truth = io.read_ground_truth_events(ground_truth)
    if isinstance(audio_durations, (str, Path)):
        audio_durations = io.read_audio_durations(audio_durations)

    if num_jobs > 1:
        from concurrent.futures import ProcessPoolExecutor, as_completed
        from functools import reduce
        from operator import add

        with ProcessPoolExecutor(max_workers=num_jobs) as executor:
            futures = [
                executor.submit(
                    _run,
                    scores, 
                    [step_filt_len],
                    merge_thresholds_abs,
                    merge_thresholds_rel,
                    either_abs_or_rel_threshold,
                )
                for step_filt_len in step_filter_lengths
            ]
            csebbs_list = [future.result() for future in as_completed(futures)]

        csebbs = reduce(add, csebbs_list)
    else:
        csebbs = []
        for step_filt_len in step_filter_lengths:
            csebbs.extend(
                _run(scores, [step_filt_len], merge_thresholds_abs, merge_thresholds_rel, either_abs_or_rel_threshold)
            )

    if folds is None:
        return selection_fn(csebbs, ground_truth, audio_durations, num_jobs=num_jobs, **selection_kwargs)
    return [selection_fn(csebbs, ground_truth, audio_durations, audio_ids=fold, num_jobs=num_jobs, **selection_kwargs) for fold in folds]


def cross_validation(
    scores: Union[str, dict],
    ground_truth: Union[str, dict],
    audio_durations: Union[str, dict],
    folds: Iterable,
    *,
    step_filter_lengths: Iterable = (0.32, 0.48, 0.64),
    merge_thresholds_abs: Iterable = (0.15, 0.2, 0.3),
    merge_thresholds_rel: Iterable = (1.5, 2.0, 3.0),
    either_abs_or_rel_threshold: bool = True,
    selection_fn: Callable,
    return_sed_scores: bool = True,
    **selection_kwargs,
) -> Tuple[list, dict, dict]:
    """perform leave-one-out cross validation. I.e., tune hyper-parameters on
    all folds but one and write outputs for the left out fold.

    Args:
        scores (str | dict of pandas.DataFrame): (path to directory with)
            SED posterior score data frames as used in sed_scores_eval
            (https://github.com/fgnt/sed_scores_eval).
        ground_truth (str | dict of list): path to/dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio as used in
            sed_scores_eval.
        audio_durations (str | dict): (path to) durations of each audio file.
        folds (iterable): optional list of audio_id sets for which best parameters
            should be selected separately in which case a list of
            CSEBBsPredictors is returned. If None, best parameters are selected
            based on the whole set and a single CSEBBsPredictor is returned.
        step_filter_lengths (iterable): candidates for step_filter_length.
        merge_thresholds_abs (iterable): candidates for merge_threshold_abs.
        merge_thresholds_rel (iterable): candidates for merge_threshold_abs.
        selection_fn (callable): function that takes list of
            (CSEBBsPredictor, Sebb predictions) tuples together with
            ground truth, audio durations, the audio ids to be considered and
            further optional arguments and selects best CSEBBsPredictor parameters.
        either_abs_or_rel_threshold (bool): if True, either merge_thresholds_rel
            or merge_thresholds_abs is used. Whereas if False, they may be used
            jointly (which, however, increases the number of parameter
            combinations that need to be evaluated during tuning).
        return_sed_scores (bool): whether to return outputs in sed_scores_eval
            format for direct use with sed_scores_eval-based evaluation.
        **selection_kwargs: any key-word arguments that should be forwarded to selection_fn.

    Returns:
        csebbs_predictors (list): list of return(s) from selection_fn
        sebbs (dict): predicted SEBBs for the whole data set,
            where predictions for different folds used different tuned hyper-parameters
        detections (dict|None): Detected events for the whole data set,
            if detection thresholds are selected by selection_fn else None.

    """
    if isinstance(scores, (str, Path)):
        scores = io.read_sed_scores(scores)
    if isinstance(ground_truth, (str, Path)):
        ground_truth = io.read_ground_truth_events(ground_truth)
    if isinstance(audio_durations, (str, Path)):
        audio_durations = io.read_audio_durations(audio_durations)
    test_ids = [set(fold) for fold in folds]
    all_ids = set.union(*test_ids)
    val_ids = [(all_ids - fold) for fold in test_ids]

    csebbs_predictors = tune(
        scores,
        ground_truth,
        audio_durations,
        folds=val_ids,
        step_filter_lengths=step_filter_lengths,
        merge_thresholds_abs=merge_thresholds_abs,
        merge_thresholds_rel=merge_thresholds_rel,
        either_abs_or_rel_threshold=either_abs_or_rel_threshold,
        selection_fn=selection_fn,
        **selection_kwargs,
    )
    sebbs = {}
    detections = {}
    for i, predictors in enumerate(csebbs_predictors):
        if not isinstance(predictors, dict):
            predictors = {"": predictors}
        for key, predictor in predictors.items():
            if isinstance(predictor, tuple):
                predictor = predictor[0]
            assert isinstance(
                predictor, CSEBBsPredictor
            ), "selection_fn doesn't seem to have returned a CSEBBsPredictor (as first return element)."
            if key not in sebbs:
                sebbs[key] = {}
                detections[key] = {}
            sebbs_i = predictor.predict(scores, audio_ids=test_ids[i])
            if predictor.detection_threshold is not None:
                detections_i = predictor.detection_thresholding(
                    sebbs_i,
                    return_sed_scores=return_sed_scores,
                )
                detections[key].update(detections_i)
            if return_sed_scores:
                sebbs_i = sed_scores_from_sebbs(sebbs_i, predictor.sound_classes)
            sebbs[key].update(sebbs_i)
    if len(sebbs) == 1 and "" in sebbs:
        sebbs = sebbs[""]
        detections = detections[""]
    return csebbs_predictors, sebbs, detections


def select_best_psds(
    csebbs: dict,
    ground_truth: dict,
    audio_durations: dict,
    audio_ids: Union[Iterable, None] = None,
    dtc_threshold: float = 0.7,
    gtc_threshold: float = 0.7,
    cttc_threshold: Union[float, None] = None,
    alpha_ct: float = 0.0,
    unit_of_time: str = "hour",
    max_efpr: float = 100.0,
    classwise: bool = True,
    num_jobs: int = 1,
) -> Tuple[CSEBBsPredictor, dict]:
    """select parameters which give highest PSDS

    Args:
        csebbs (list): list of (CSEBBsPredictor, SEBB predictions) from grid search performed in tune function above
        ground_truth (dict of list): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio as used in
            sed_scores_eval.
        audio_durations: The duration of each audio file.
        audio_ids (iterable|None): audio ids that parameter selection should be restricted to.
        dtc_threshold (float): detection tolerance criterion threshold of PSDS
        gtc_threshold (float): ground truth intersection criterion threshold of PSDS
        cttc_threshold (float|None): cross trigger tolerance criterion threshold of PSDS
        alpha_ct (float): cross trigger penalization weight of PSDS.
        unit_of_time (str): time unit for FPR computation in PSDS
        max_efpr (float): maximum false positives per time unit to be considered in PSD-ROC.
        classwise (bool): whether to use different parameters for each sound_class or not.

    Returns:
        csebbs_predictor: CSEBBsPredictor with best hyper-parameters w.r.t. PSDS
        best_values: class-wise PSDS values with best hyper-parameters

    """
    if audio_ids is not None:
        csebbs = [
            (predictor, {audio_id: csebbs_i[audio_id] for audio_id in audio_ids}) for predictor, csebbs_i in csebbs
        ]
        ground_truth = {audio_id: ground_truth[audio_id] for audio_id in audio_ids}
        audio_durations = {audio_id: audio_durations[audio_id] for audio_id in audio_ids}
    # convert sebbs to sed_scores_eval score format for PSDS evaluation.
    csebbs_scores = [
        (
            predictor,
            sed_scores_from_sebbs(csebbs_i, sound_classes=predictor.sound_classes, audio_duration=audio_durations),
        )
        for predictor, csebbs_i in tqdm(csebbs, desc="Compute sed_scores", total=len(csebbs))
    ]
    best_step_filter_length = {}
    best_merge_threshold_abs = {}
    best_merge_threshold_rel = {}
    best_values = {}
    for predictor, csebbs_scores_i in tqdm(csebbs_scores, desc="Select best PSDS", total=len(csebbs_scores)):
        single_class_psds = intersection_based.psds(
            scores=csebbs_scores_i,
            ground_truth=ground_truth,
            audio_durations=audio_durations,
            dtc_threshold=dtc_threshold,
            gtc_threshold=gtc_threshold,
            cttc_threshold=cttc_threshold,
            alpha_ct=alpha_ct,
            unit_of_time=unit_of_time,
            max_efpr=max_efpr,
            num_jobs=num_jobs,
        )[1]
        mean = np.mean(list(single_class_psds.values()))
        for sound_class in single_class_psds:
            value = single_class_psds[sound_class] if classwise else mean
            if (sound_class not in best_values) or (value > best_values[sound_class]):
                # for this sound class the performance achieved with current params is better than with previous params
                best_values[sound_class] = value
                best_step_filter_length[sound_class] = predictor.step_filter_length
                best_merge_threshold_abs[sound_class] = predictor.merge_threshold_abs
                best_merge_threshold_rel[sound_class] = predictor.merge_threshold_rel

    csebbs_predictor = CSEBBsPredictor(
        step_filter_length=best_step_filter_length,
        merge_threshold_rel=best_merge_threshold_rel,
        merge_threshold_abs=best_merge_threshold_abs,
    )
    return csebbs_predictor, best_values


def select_best_cbf(
    csebbs: dict,
    ground_truth: dict,
    audio_durations: dict,
    audio_ids: Union[Iterable, None] = None,
    onset_collar: float = 0.2,
    offset_collar: float = 0.2,
    offset_collar_rate: float = 0.2,
    classwise: bool = True,
    num_jobs: int = 1,
    **kwargs,
) -> Tuple[CSEBBsPredictor, dict]:
    """select parameters which give highest collar-based F1-score

    Args:
        csebbs (list): list of (CSEBBsPredictor, SEBB predictions) from grid search performed in tune function above
        ground_truth (dict of list): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio as used in
            sed_scores_eval.
        audio_durations: The duration of each audio file.
        audio_ids (iterable|None): audio ids that parameter selection should be restricted to.
        onset_collar (float): onset collar in seconds to be used in collar-based evaluation.
        offset_collar (float): min offset collar in seconds to be used in collar-based evaluation.
        offset_collar_rate (float): min offset collar as a rate of the ground truth event length
            to be used in collar-based evaluation. Actual offset collar given as
            offset_collar_for_gt_event = max(
                offset_collar, offset_collar_rate*length_of_gt_event_in_seconds
            )
        classwise (bool): whether to use different parameters for each sound_class or not.

    Returns:
        csebbs_predictor: CSEBBsPredictor with best hyper-parameters w.r.t. collar-based F1-score
        best_values: class-wise collar-based F1-score values with best hyper-params

    """
    if audio_ids is not None:
        csebbs = [
            (predictor, {audio_id: csebbs_i[audio_id] for audio_id in audio_ids}) for predictor, csebbs_i in csebbs
        ]
        ground_truth = {audio_id: ground_truth[audio_id] for audio_id in audio_ids}
        audio_durations = {audio_id: audio_durations[audio_id] for audio_id in audio_ids}
    # convert sebbs to sed_scores_eval score format for F1-score evaluation.
    csebbs_scores = [
        (
            predictor,
            sed_scores_from_sebbs(csebbs_i, sound_classes=predictor.sound_classes, audio_duration=audio_durations),
        )
        for predictor, csebbs_i in tqdm(csebbs, desc="Compute sed_scores", total=len(csebbs))
    ]
    best_step_filter_length = {}
    best_merge_threshold_abs = {}
    best_merge_threshold_rel = {}
    best_detection_threshold = {}
    best_values = {}

    for predictor, csebbs_scores_i in tqdm(csebbs_scores, desc="Select best CBF", total=len(csebbs_scores)):
        f, _, _, thresholds_cbf, _ = collar_based.best_fscore(
            scores=csebbs_scores_i,
            ground_truth=ground_truth,
            onset_collar=onset_collar,
            offset_collar=offset_collar,
            offset_collar_rate=offset_collar_rate,
            num_jobs=num_jobs,
        )
        for sound_class in thresholds_cbf:
            value = f[sound_class] if classwise else f["macro_average"]
            if (sound_class not in best_values) or (value > best_values[sound_class]):
                # for this sound class the performance achieved with current params is better than with previous params
                best_values[sound_class] = value
                best_step_filter_length[sound_class] = predictor.step_filter_length
                best_merge_threshold_abs[sound_class] = predictor.merge_threshold_abs
                best_merge_threshold_rel[sound_class] = predictor.merge_threshold_rel
                best_detection_threshold[sound_class] = thresholds_cbf[sound_class]

    csebbs_predictor = CSEBBsPredictor(
        step_filter_length=best_step_filter_length,
        merge_threshold_rel=best_merge_threshold_rel,
        merge_threshold_abs=best_merge_threshold_abs,
        detection_threshold=best_detection_threshold,
    )
    return csebbs_predictor, best_values


def select_best_psds_and_cbf(
    csebbs: list,
    ground_truth: dict,
    audio_durations: dict,
    audio_ids: Union[Iterable, None] = None,
    dtc_threshold: float = 0.7,
    gtc_threshold: float = 0.7,
    cttc_threshold: Union[float, None] = None,
    alpha_ct: float = 0.0,
    unit_of_time: str = "hour",
    max_efpr: float = 100.0,
    onset_collar: float = 0.2,
    offset_collar: float = 0.2,
    offset_collar_rate: float = 0.2,
    classwise: bool = True,
    num_jobs: int = 1,
) -> dict:
    """Select both, best PSDS and collar-based parameters

    Args:
        csebbs (list): list of (CSEBBsPredictor, SEBB predictions) from grid search performed in tune function above
        ground_truth (dict of list): dict of lists of ground truth
            event tuples (onset, offset, event label) for each audio as used in
            sed_scores_eval.
        audio_durations: The duration of each audio file.
        audio_ids (iterable|None): audio ids that parameter selection should be restricted to.
        dtc_threshold (float): detection tolerance criterion threshold of PSDS
        gtc_threshold (float): ground truth intersection criterion threshold of PSDS
        cttc_threshold (float|None): cross trigger tolerance criterion threshold of PSDS
        alpha_ct (float): cross trigger penalization weight of PSDS.
        unit_of_time (str): time unit for FPR computation in PSDS
        max_efpr (float): maximum false positives per time unit to be considered in PSD-ROC.
        onset_collar (float): onset collar in seconds to be used in collar-based evaluation.
        offset_collar (float): min offset collar in seconds to be used in collar-based evaluation.
        offset_collar_rate (float): min offset collar as a rate of the ground truth event length
            to be used in collar-based evaluation. Actual offset collar given as
            offset_collar_for_gt_event = max(
                offset_collar, offset_collar_rate*length_of_gt_event_in_seconds
            )
        classwise (bool): whether to use different parameters for each sound_class or not.

    Returns: dict of (tuned CSEBBsPredictors, class-wise performance dict) tuples
        for both PSDS and collar-based F1-score metrics

    """
    return {
        "psds": select_best_psds(
            csebbs,
            ground_truth,
            audio_durations,
            audio_ids=audio_ids,
            dtc_threshold=dtc_threshold,
            gtc_threshold=gtc_threshold,
            cttc_threshold=cttc_threshold,
            alpha_ct=alpha_ct,
            unit_of_time=unit_of_time,
            max_efpr=max_efpr,
            classwise=classwise,
            num_jobs=num_jobs,
        ),
        "cbf": select_best_cbf(
            csebbs,
            ground_truth,
            audio_durations,
            audio_ids=audio_ids,
            onset_collar=onset_collar,
            offset_collar=offset_collar,
            offset_collar_rate=offset_collar_rate,
            classwise=classwise,
            num_jobs=num_jobs,
        ),
    }



# Copyright (C) 2024 Mitsubishi Electric Research Laboratories (MERL)
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import numpy as np
from sed_scores_eval import collar_based, intersection_based, io
from sed_scores_eval.base_modules.scores import create_score_dataframe

from sebbs import csebbs, package_dir


def test_csebbs_simple():
    # paper example:
    cos = (1 - np.cos(np.linspace(0, 2 * np.pi, 270))) / 2
    y = np.zeros(500)
    y[15 : 15 + len(cos)] += 0.98 * cos
    cos = (1 - np.cos(np.linspace(0, 2 * np.pi, 170))) / 2
    y[315 : 315 + len(cos)] += 0.2 * cos
    timestamps = np.linspace(0.0, 10.0, len(y) + 1)
    scores_df = create_score_dataframe(y[:, None], timestamps, event_classes=["a"])
    sebbs_pred = csebbs.CSEBBsPredictor(
        step_filter_length=1.0,
        merge_threshold_abs=1.0,
        merge_threshold_rel=2.0,
    ).predict({"0": scores_df})
    sebbs_onsets, sebbs_offsets, *_ = list(zip(*sebbs_pred["0"]))
    assert sebbs_onsets == (1.66, 7.16), sebbs_onsets
    assert sebbs_offsets == (4.34, 8.84), sebbs_offsets


def test_csebbs_oracle_tuning_psds():
    # load predictions, ground_truth, and audio_durations
    scores = io.read_sed_scores(package_dir / "tests" / "dcase2023task4a" / "baseline2_run1_scores")
    ground_truth = io.read_ground_truth_events(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "ground_truth.tsv"
    )
    ground_truth = {audio_id: ground_truth[audio_id] for audio_id in ground_truth if audio_id in scores}
    scores = {audio_id: scores[audio_id] for audio_id in ground_truth}
    audio_durations = io.read_audio_durations(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "audio_durations.tsv"
    )
    audio_durations = {audio_id: audio_durations[audio_id] for audio_id in scores}

    # tune hyper-parameters on subset of DESED's public eval
    csebbs_predictor, best_psds_values = csebbs.tune(
        scores=scores,
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        step_filter_lengths=(0.32, 0.48, 0.64),
        merge_thresholds_abs=(0.15, 0.2, 0.3),
        merge_thresholds_rel=(1.5, 2.0, 3.0),
        selection_fn=csebbs.select_best_psds,
    )

    # run inference on subset of DESED's public eval
    sed_scores = csebbs_predictor.predict(scores, return_sed_scores=True)

    # evaluate PSDS performance
    psds, single_class_psds, *_ = intersection_based.psds(
        scores=sed_scores,
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        dtc_threshold=0.7,
        gtc_threshold=0.7,
        alpha_ct=0.0,
        alpha_st=1.0,
        unit_of_time="hour",
        max_efpr=100.0,
    )
    for sound_class in single_class_psds:
        assert abs(single_class_psds[sound_class] - best_psds_values[sound_class]) < 1e-6, (
            sound_class,
            single_class_psds[sound_class],
            best_psds_values[sound_class],
        )


def test_csebbs_oracle_tuning_psds_and_cbf():
    # load predictions, ground_truth, and audio_durations
    scores = io.read_sed_scores(package_dir / "tests" / "dcase2023task4a" / "baseline2_run1_scores")
    ground_truth = io.read_ground_truth_events(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "ground_truth.tsv"
    )
    ground_truth = {audio_id: ground_truth[audio_id] for audio_id in ground_truth if audio_id in scores}
    scores = {audio_id: scores[audio_id] for audio_id in ground_truth}
    audio_durations = io.read_audio_durations(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "audio_durations.tsv"
    )
    audio_durations = {audio_id: audio_durations[audio_id] for audio_id in scores}

    # tune hyper-parameters on subset of DESED's public eval
    csebbs_predictors = csebbs.tune(
        scores=scores,
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        step_filter_lengths=(0.32, 0.48, 0.64),
        merge_thresholds_abs=(0.15, 0.2, 0.3),
        merge_thresholds_rel=(1.5, 2.0, 3.0),
        selection_fn=csebbs.select_best_psds_and_cbf,
    )

    # psds predictor
    predictor_psds, best_psds_values = csebbs_predictors["psds"]

    # run inference on subset of DESED's public eval
    sed_scores = predictor_psds.predict(scores, return_sed_scores=True)

    # evaluate PSDS performance
    psds, single_class_psds, *_ = intersection_based.psds(
        scores=sed_scores,
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        dtc_threshold=0.7,
        gtc_threshold=0.7,
        alpha_ct=0.0,
        alpha_st=1.0,
        unit_of_time="hour",
        max_efpr=100.0,
    )
    for sound_class in single_class_psds:
        assert abs(single_class_psds[sound_class] - best_psds_values[sound_class]) < 1e-6, (
            sound_class,
            single_class_psds[sound_class],
            best_psds_values[sound_class],
        )

    # cbf predictor
    predictor_cbf, best_cbf_values = csebbs_predictors["cbf"]

    # the following two blocks should give the same results:
    # 1) soft score inference and thresholding in collar_based.fscore
    sed_scores = predictor_cbf.predict(scores, return_sed_scores=True)
    f, *_ = collar_based.fscore(
        scores=sed_scores,
        ground_truth=ground_truth,
        threshold=predictor_cbf.detection_threshold,
        onset_collar=0.2,
        offset_collar=0.2,
        offset_collar_rate=0.2,
    )
    for sound_class in predictor_cbf.sound_classes:
        assert abs(f[sound_class] - best_cbf_values[sound_class]) < 1e-6, (
            sound_class,
            f[sound_class],
            best_cbf_values[sound_class],
        )

    # 2) binary score inference (detection) and simple 0.5 threshold in collar_based.fscore
    sed_scores = predictor_cbf.detect(scores, return_sed_scores=True)
    f, *_ = collar_based.fscore(
        scores=sed_scores,
        ground_truth=ground_truth,
        threshold=0.5,  # can be any number between 0 and 1 here as the sed_scores obtained from detection are binary
        onset_collar=0.2,
        offset_collar=0.2,
        offset_collar_rate=0.2,
    )
    for sound_class in predictor_cbf.sound_classes:
        assert abs(f[sound_class] - best_cbf_values[sound_class]) < 1e-6, (
            sound_class,
            f[sound_class],
            best_cbf_values[sound_class],
        )


def test_csebbs_cross_validation_psds():
    # load predictions, ground_truth, and audio_durations
    scores = io.read_sed_scores(package_dir / "tests" / "dcase2023task4a" / "baseline2_run1_scores")
    ground_truth = io.read_ground_truth_events(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "ground_truth.tsv"
    )
    ground_truth = {audio_id: ground_truth[audio_id] for audio_id in ground_truth if audio_id in scores}
    scores = {audio_id: scores[audio_id] for audio_id in ground_truth}
    audio_durations = io.read_audio_durations(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "audio_durations.tsv"
    )
    audio_durations = {audio_id: audio_durations[audio_id] for audio_id in scores}

    # 5-fold cross validation split
    n_folds = 5
    all_keys = scores.keys()
    shuffled_keys = sorted(all_keys)
    np.random.RandomState(0).shuffle(shuffled_keys)
    split_idx = np.linspace(0, len(shuffled_keys), n_folds + 1).astype(int).tolist()
    folds = [set(shuffled_keys[split_idx[i] : split_idx[i + 1]]) for i in range(n_folds)]

    # cross validation
    csebbs_predictors, sed_scores, detections = csebbs.cross_validation(
        scores=scores,
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        folds=folds,
        step_filter_lengths=(0.32, 0.48, 0.64),
        merge_thresholds_abs=(0.15, 0.2, 0.3),
        merge_thresholds_rel=(1.5, 2.0, 3.0),
        selection_fn=csebbs.select_best_psds,
        return_sed_scores=True,
    )

    # evaluate PSDS performance
    psds, single_class_psds, *_ = intersection_based.psds(
        scores=sed_scores,
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        dtc_threshold=0.7,
        gtc_threshold=0.7,
        alpha_ct=0.0,
        alpha_st=1.0,
        unit_of_time="hour",
        max_efpr=100.0,
    )
    # comment assert as evaluating reduced set (20 files) only
    # assert abs(psds - 0.6340242849210951) < 1e-6, psds  # result from paper


def test_csebbs_cross_validation_psds_cbf():
    # load predictions, ground_truth, and audio_durations
    scores = io.read_sed_scores(package_dir / "tests" / "dcase2023task4a" / "baseline2_run1_scores")
    ground_truth = io.read_ground_truth_events(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "ground_truth.tsv"
    )
    ground_truth = {audio_id: ground_truth[audio_id] for audio_id in ground_truth if audio_id in scores}
    scores = {audio_id: scores[audio_id] for audio_id in ground_truth}
    audio_durations = io.read_audio_durations(
        package_dir / "tests" / "dcase2023task4a" / "metadata" / "audio_durations.tsv"
    )
    audio_durations = {audio_id: audio_durations[audio_id] for audio_id in ground_truth}

    # 5-fold cross validation split
    n_folds = 5
    all_keys = scores.keys()
    shuffled_keys = sorted(all_keys)
    np.random.RandomState(0).shuffle(shuffled_keys)
    split_idx = np.linspace(0, len(shuffled_keys), n_folds + 1).astype(int).tolist()
    folds = [set(shuffled_keys[split_idx[i] : split_idx[i + 1]]) for i in range(n_folds)]

    # cross validation
    csebbs_predictors, sed_scores, detections = csebbs.cross_validation(
        scores=scores,
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        folds=folds,
        step_filter_lengths=(0.32, 0.48, 0.64),
        merge_thresholds_abs=(0.15, 0.2, 0.3),
        merge_thresholds_rel=(1.5, 2.0, 3.0),
        selection_fn=csebbs.select_best_psds_and_cbf,
        return_sed_scores=True,
    )

    # evaluate PSDS performance
    psds, single_class_psds, *_ = intersection_based.psds(
        scores=sed_scores["psds"],
        ground_truth=ground_truth,
        audio_durations=audio_durations,
        dtc_threshold=0.7,
        gtc_threshold=0.7,
        alpha_ct=0.0,
        alpha_st=1.0,
        unit_of_time="hour",
        max_efpr=100.0,
    )
    # comment assert as evaluating reduced set (20 files) only
    # assert abs(psds - 0.6340242849210951) < 1e-6, psds  # result from paper

    # evaluate collar-based F1-score performance
    f, *_ = collar_based.fscore(
        scores=detections["cbf"],
        ground_truth=ground_truth,
        threshold=0.5,  # can be any number between 0 and 1 here as the sed_scores obtained from detection are binary
        onset_collar=0.2,
        offset_collar=0.2,
        offset_collar_rate=0.2,
    )
    # comment assert as evaluating reduced set (20 files) only
    # assert abs(f['macro_average'] - 0.6721148195872252) < 1e-6, f['macro_average']  # result from paper



# Copyright (C) 2024 Mitsubishi Electric Research Laboratories (MERL)
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import numpy as np
from sed_scores_eval.base_modules.scores import create_score_dataframe
from sed_scores_eval.utils.array_ops import get_first_index_where


def sed_scores_from_sebbs(sebbs, sound_classes, audio_duration=None, fill_value=-np.inf):
    """Save Sebbs in sed_scores_eval format with constant class score over the extent of a SEBB and -inf else.

    Args:
        sebbs (dict | list): (dict of) list of SEBBs (onset, offset, class, confidence) as provided by predict method.
        sound_classes: list of all sound classes.
        audio_duration: optional (dict of) audio duration(s).
        fill_value

    Returns:

    """
    if isinstance(sebbs, dict):
        return {
            key: sed_scores_from_sebbs(
                preds_for_key,
                sound_classes,
                audio_duration[key] if audio_duration is not None else None,
                fill_value=fill_value,
            )
            for key, preds_for_key in sebbs.items()
        }
    if len(sebbs) == 0:
        return create_score_dataframe(
            np.full((1, len(sound_classes)), fill_value),
            [0.0, 0.001 if audio_duration is None else audio_duration],
            sound_classes,
        )
    change_points = set()
    for bb in sebbs:
        change_points.update([bb[0], bb[1]])
    change_points = sorted(change_points)
    if len(change_points) == 0 or change_points[0] != 0.0:
        change_points = [0.0] + change_points
    if audio_duration is not None and change_points[-1] != audio_duration:
        change_points = change_points + [audio_duration]
    change_points = np.unique(change_points)
    score_arr = np.full((len(change_points) - 1, len(sound_classes)), fill_value)
    for cand_onset, cand_offset, class_name, mean_score in sebbs:
        cand_onset_idx = get_first_index_where(change_points, "geq", cand_onset)
        cand_offset_idx = get_first_index_where(change_points, "geq", cand_offset)
        class_idx = sound_classes.index(class_name)
        score_arr[cand_onset_idx:cand_offset_idx, class_idx] = mean_score
    return create_score_dataframe(score_arr, change_points, sound_classes)


def sed_scores_from_detections(detections, sound_classes, audio_duration=None, fill_value=-np.inf):
    return sed_scores_from_sebbs(
        {audio_id: [(*event, 1.0) for event in events] for audio_id, events in detections.items()},
        sound_classes=sound_classes,
        audio_duration=audio_duration,
        fill_value=fill_value,
    )



# Copyright (C) 2024 Mitsubishi Electric Research Laboratories (MERL)
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from codecs import open
from os import path

import numpy as np
import setuptools as st
from Cython.Build import cythonize

ext_modules = cythonize(
    [
        "sebbs/change_detection.pyx",
    ]
)

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

st.setup(
    name="sebbs",
    version="1.0.1",
    description="Sound Event Bounding Boxes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/merlresearch/sound_event_bounding_boxes",
    author="Mitsubishi Electric Research Laboratories",
    author_email="ebbers[at]merl[dot]com",
    license="AGPL-3.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        # "License :: OSI Approved :: AGPL-3.0 License",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
    keywords="sound event detection, polyphonic sound detection, post processing, change detection",
    packages=st.find_packages(exclude=["contrib", "docs", "tests*"]),
    install_requires=[
        "numpy",
        "pandas",
        "scipy",
        "pathlib",
        "sed_scores_eval>=0.0.4",
        "Cython",
    ],
    extras_require={
        "dev": ["check-manifest"],
        "test": ["coverage", "jupyter", "matplotlib"],
    },
    ext_modules=ext_modules,
    package_data={"sebbs": ["**/*.pyx"]},  # https://stackoverflow.com/a/60751886
    include_dirs=[np.get_include()],
)
